# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli Dbrx Checkpoint Converter."""


from __future__ import annotations

from typing import cast

from transformers.models.dbrx.configuration_dbrx import (  # type: ignore[import]
    DbrxConfig,
    DbrxFFNConfig,
)

from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.modules.converter.base import FP8OnlyConverter
from friendli.modules.converter.interface import RotaryEmbeddingConversionInterface


class DbrxForCausalLMConverter(FP8OnlyConverter, RotaryEmbeddingConversionInterface):
    """DbrxForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if Dbrx architectures' config can be converted to Friendli format."""
        super().check_config()
        config = cast(DbrxConfig, self.config)
        try:
            if config.tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=True'",
                    valid_options=[False],
                )
            if config.ffn_config.moe_top_k not in [1, 2, 4]:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'moe_top_k={config.ffn_config.moe_top_k}'",
                    valid_options=[1, 2, 4],
                )
            if config.ffn_config.moe_num_experts not in [1, 2, 4, 8, 16]:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'moe_num_experts={config.ffn_config.moe_num_experts}'",
                    valid_options=[1, 2, 4, 8, 16],
                )

        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    @property
    def model_type(self) -> str:
        """Model type."""
        return "dbrx"

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before LLaMA's transformer block number."""
        return "transformer.blocks."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in LLaMA."""
        return cast(DbrxConfig, self.config).n_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in LLaMA."""
        return cast(DbrxConfig, self.config).d_model

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in LLaMA."""
        return cast(DbrxConfig, self.config).n_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in LLaMA."""
        config = cast(DbrxConfig, self.config)
        if config.attn_config.kv_n_heads is None:
            return self.decoder_num_attention_heads
        return config.attn_config.kv_n_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of LLaMA."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate size of the linear layer in LLaMA MLP."""
        return cast(DbrxConfig, self.config).ffn_config.ffn_hidden_size

    @property
    def rotary_dim(self) -> int:
        """The rotary embedding dimension of LLaMA."""
        return self.decoder_head_size

    @property
    def rotary_emb_base(self) -> float:
        """The rotary embedding base of LLaMA."""
        return cast(DbrxConfig, self.config).attn_config.rope_theta



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Quantizer Maps."""

from __future__ import annotations

from typing import Any, Dict, Type

from friendli.enums import QuantMode
from friendli.errors import NotSupportedQuantModeError
from friendli.modules.converter.base import OneOfConverter
from friendli.modules.converter.utils import get_model_arch
from friendli.modules.quantizer.awq.base import AWQHook, AWQQuantizer
from friendli.modules.quantizer.awq.models.gpt_neox import AWQGPTNeoXHook
from friendli.modules.quantizer.awq.models.gptj import AWQGPTJHook
from friendli.modules.quantizer.awq.models.llama import AWQLlamaHook
from friendli.modules.quantizer.awq.models.mpt import AWQMPTHook
from friendli.modules.quantizer.base import CommonQuantizer, FP8QuantHook, FP8Quantizer
from friendli.modules.quantizer.models.dbrx import DbrxHook
from friendli.modules.quantizer.models.llama import LlamaHook
from friendli.modules.quantizer.models.mixtral import MixtralHook
from friendli.modules.quantizer.models.mpt import MPTHook
from friendli.modules.quantizer.models.phi3 import Phi3Hook
from friendli.modules.quantizer.schema.config import OneOfQuantConfig
from friendli.modules.quantizer.smoothquant.base import (
    SmoothQuantHook,
    SmoothQuantQuantizer,
)
from friendli.modules.quantizer.smoothquant.models.bloom import SmoothQuantBloomHook
from friendli.modules.quantizer.smoothquant.models.codegen import SmoothQuantCodeGenHook
from friendli.modules.quantizer.smoothquant.models.falcon import SmoothQuantFalconHook
from friendli.modules.quantizer.smoothquant.models.gpt2 import SmoothQuantGPT2Hook
from friendli.modules.quantizer.smoothquant.models.gpt_neox import (
    SmoothQuantGPTNeoXHook,
)
from friendli.modules.quantizer.smoothquant.models.gptj import SmoothQuantGPTJHook
from friendli.modules.quantizer.smoothquant.models.llama import SmoothQuantLlamaHook
from friendli.modules.quantizer.smoothquant.models.mpt import SmoothQuantMPTHook
from friendli.modules.quantizer.smoothquant.models.opt import SmoothQuantOPTHook

model_arch_smoothquant_hook_map: Dict[str, type[SmoothQuantHook]] = {
    "OPTForCausalLM": SmoothQuantOPTHook,
    "MPTForCausalLM": SmoothQuantMPTHook,
    "BloomForCausalLM": SmoothQuantBloomHook,
    "CodeGenForCausalLM": SmoothQuantCodeGenHook,
    "GPTNeoXForCausalLM": SmoothQuantGPTNeoXHook,
    "GPTJForCausalLM": SmoothQuantGPTJHook,
    "GPT2LMHeadModel": SmoothQuantGPT2Hook,
    "FalconForCausalLM": SmoothQuantFalconHook,
    "LlamaForCausalLM": SmoothQuantLlamaHook,
}

model_arch_awq_hook_map: Dict[str, type[AWQHook]] = {
    "GPTJForCausalLM": AWQGPTJHook,
    "GPTNeoXForCausalLM": AWQGPTNeoXHook,
    "LlamaForCausalLM": AWQLlamaHook,
    "MPTForCausalLM": AWQMPTHook,
    "MistralForCausalLM": AWQLlamaHook,
}

model_arch_fp8_hook_map: Dict[str, type[FP8QuantHook]] = {
    "LlamaForCausalLM": LlamaHook,
    "MistralForCausalLM": LlamaHook,
    "MixtralForCausalLM": MixtralHook,
    "MPTForCausalLM": MPTHook,
    "CohereForCausalLM": LlamaHook,
    "DbrxForCausalLM": DbrxHook,
    "Phi3ForCausalLM": Phi3Hook,
}


def get_quanthook_map(quant_mode: QuantMode) -> Dict[str, Any]:
    """Get quantizer map."""
    if quant_mode == QuantMode.SMOOTH_QUANT:
        return model_arch_smoothquant_hook_map
    if quant_mode == QuantMode.AWQ:
        return model_arch_awq_hook_map
    if quant_mode == QuantMode.FP8:
        return model_arch_fp8_hook_map
    raise NotSupportedQuantModeError(
        invalid_option=quant_mode,
        valid_options=[e.value for e in QuantMode],
    )


def get_quantizer_class(quant_mode: QuantMode) -> Type[CommonQuantizer]:
    """Get quantizer class."""
    if quant_mode == QuantMode.SMOOTH_QUANT:
        return SmoothQuantQuantizer
    if quant_mode == QuantMode.AWQ:
        return AWQQuantizer
    if quant_mode == QuantMode.FP8:
        return FP8Quantizer
    raise NotSupportedQuantModeError(
        invalid_option=quant_mode,
        valid_options=[e.value for e in QuantMode],
    )


def get_quantized_converter(
    quant_config: OneOfQuantConfig,
    converter: OneOfConverter,
) -> CommonQuantizer:
    """Get quantizer for specific model architecture with quant mode and args."""
    model_arch = get_model_arch(converter.config)
    quant_mode = quant_config.mode
    quantizer = get_quantizer_class(quant_mode)
    quanthook_map = get_quanthook_map(quant_mode)
    quanthook = quanthook_map[model_arch](quant_config, converter)
    return quantizer(quanthook, quant_config, converter)



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

"""Savers to save a converted checkpoints into various file types."""

from __future__ import annotations

import json
import os
from abc import abstractmethod
from contextlib import AbstractContextManager
from typing import Dict, Generic, List, TypeVar, Union

import h5py  # type: ignore[import]
import numpy as np
import safetensors.numpy  # type: ignore[import]
import safetensors.torch  # type: ignore[import]
import torch
from typing_extensions import Self

from friendli.enums import CheckpointFileType
from friendli.errors import CheckpointConversionError
from friendli.logging import logger


def get_saver(
    ckpt_file_type: CheckpointFileType, output_dir: str, output_file_name: str
) -> CheckpointSaver:
    """Create a saver that corresponds to the file type."""
    if ckpt_file_type == CheckpointFileType.HDF5:
        return HDF5Saver(output_dir, output_file_name)
    if ckpt_file_type == CheckpointFileType.SAFETENSORS:
        return SafetensorsSaver(output_dir, output_file_name)
    raise CheckpointConversionError(
        f"Output file type {ckpt_file_type} is not supported."
    )


class CheckpointSaver(AbstractContextManager):
    """Abstract for savers."""

    def __init__(
        self, output_dir: Union[str, os.PathLike], output_file_name: str
    ) -> None:
        """Check that the output file already exists."""
        super().__init__()
        self._output_dir = output_dir
        self._output_file_name = output_file_name

    @abstractmethod
    def save_tensor(self, tensor_id: str, t: Union[np.ndarray, torch.Tensor]) -> None:
        """Save the tensor in the file."""
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """Close the output checkpoint file."""
        raise NotImplementedError

    def __enter__(self) -> Self:
        """Enter for context manager."""
        return self

    def __exit__(self, *exc) -> None:
        """Exit for context manager."""
        self.close()


class HDF5Saver(CheckpointSaver):
    """Saver for HDF5."""

    def __init__(self, output_dir: str, output_file_name: str) -> None:
        """Create a HDF5 file."""
        super().__init__(output_dir, output_file_name)
        self._out_f = h5py.File(os.path.join(output_dir, output_file_name), "w")

    def save_tensor(self, tensor_id: str, t: Union[np.ndarray, torch.Tensor]) -> None:
        """Create a group if not exists, and save the tensor in the file."""
        assert isinstance(t, np.ndarray)
        self._out_f[tensor_id] = t

    def close(self) -> None:
        """Close the HDF5 file."""
        self._out_f.close()


T = TypeVar("T")


class SafetensorsSaverInterface(Generic[T]):
    """Interface for saving safetensor format."""

    def get_weight_size(self, tensor: T) -> int:
        """Get total weight size in `Byte` unit."""
        raise NotImplementedError

    def save_file(self, tensor: Dict[str, T], path: str) -> None:
        """Save given tensor to path."""
        raise NotImplementedError


class TorchSafetensorsSaverInterface(SafetensorsSaverInterface[torch.Tensor]):
    """Interface for saving safetensor format."""

    def get_weight_size(self, tensor: torch.Tensor) -> int:
        """Get total weight size in `Byte` unit."""
        return tensor.itemsize * tensor.numel()

    def save_file(self, tensor: Dict[str, torch.Tensor], path: str) -> None:
        """Save given tensor to path."""
        safetensors.torch.save_file(tensor, path)


class NumpySafetensorsSaverInterface(SafetensorsSaverInterface[np.ndarray]):
    """Interface for saving safetensor format."""

    def get_weight_size(self, tensor: np.ndarray) -> int:
        """Get total weight size in `Byte` unit."""
        return tensor.itemsize * tensor.size

    def save_file(self, tensor: Dict[str, np.ndarray], path: str) -> None:
        """Save given tensor to path."""
        safetensors.numpy.save_file(tensor, path)


class UnionSafetensorsSaverInterface(
    SafetensorsSaverInterface[Union[torch.Tensor, np.ndarray]]
):
    """Interface for saving safetensor format."""

    def __init__(self) -> None:
        """Initialize UnionSafetensorsSaverInterface."""
        self._sub_itfcs = {
            np.ndarray: NumpySafetensorsSaverInterface(),
            torch.Tensor: TorchSafetensorsSaverInterface(),
        }
        super().__init__()

    def get_weight_size(self, tensor: Union[torch.Tensor, np.ndarray]) -> int:
        """Get total weight size in `Byte` unit."""
        return self._sub_itfcs[type(tensor)].get_weight_size(tensor)  # type: ignore[attr-defined]

    def save_file(
        self, tensor: Dict[str, Union[torch.Tensor, np.ndarray]], path: str
    ) -> None:
        """Save given tensor to path."""
        if len(tensor) == 0:
            logger.warn("No tensor to save. Skip saving tensors..")
            return
        # NOTE: Assume that all tensors are the same type
        tensor_type = type(next(iter(tensor.values())))
        itfc = self._sub_itfcs[tensor_type]
        itfc.save_file(tensor, path)  # type: ignore[attr-defined]


class SafetensorsSaver(CheckpointSaver):
    """Saver for Safetensors.

    This temporally saves the converted tensors in local memory.
    Then, all of the tensors are saved in the file at a time when close() is called,
    because Safetensors does not support stream saving.
    """

    def __init__(
        self, output_dir: Union[str, os.PathLike], output_file_name: str
    ) -> None:
        """Initialize a saver."""
        super().__init__(output_dir, output_file_name)
        self._tensors: Dict[str, Union[np.ndarray, torch.Tensor]] = {}
        self._saver: UnionSafetensorsSaverInterface = UnionSafetensorsSaverInterface()

    def save_tensor(self, tensor_id: str, t: Union[np.ndarray, torch.Tensor]) -> None:
        """Save the tensor in the local memory."""
        self._tensors[tensor_id] = t

    def shard_checkpoint(self, max_shard_size: str):
        """Shard the checkpoint with index."""
        # pylint: disable=too-many-locals
        int_max_shard_size = int(max_shard_size[:-2]) * (10**9)
        sharded_tensors: List[Dict[str, Union[np.ndarray, torch.Tensor]]] = [{}]
        last_block_size = 0
        total_size = 0

        for key, weight in self._tensors.items():
            weight_size = self._saver.get_weight_size(weight)
            if (
                last_block_size + weight_size > int_max_shard_size
                and len(sharded_tensors[-1]) > 0
            ):
                sharded_tensors.append({})
                last_block_size = 0

            sharded_tensors[-1][key] = weight
            last_block_size += weight_size
            total_size += weight_size

        if len(sharded_tensors) == 1:
            return {self._output_file_name: sharded_tensors[0]}, None

        weight_map = {}
        shards = {}
        for idx, shard in enumerate(sharded_tensors):
            shard_file = self._output_file_name.replace(
                ".safetensors",
                f"-{idx + 1:05d}-of-{len(sharded_tensors):05d}.safetensors",
            )
            shards[shard_file] = shard
            for key in shard.keys():
                weight_map[key] = shard_file

        metadata = {"total_size": total_size}
        index = {"metadata": metadata, "weight_map": weight_map}
        return shards, index

    def _save_to_file(self) -> None:
        """Save the tensors in the file."""
        logger.info("Saving the converted checkpoint...")

        max_shard_size = "10GB"
        shards, index = self.shard_checkpoint(max_shard_size)

        for shard_file, shard in shards.items():
            self._saver.save_file(shard, os.path.join(self._output_dir, shard_file))

        if index is None:
            path_to_weights = os.path.join(self._output_dir, self._output_file_name)
            logger.info("Model weights saved in (%s)", path_to_weights)
        else:
            save_index_file = os.path.join(
                self._output_dir, "model.safetensors.index.json"
            )
            # Save the index as well
            with open(save_index_file, "w", encoding="utf-8") as f:
                content = json.dumps(index, indent=2, sort_keys=True) + "\n"
                f.write(content)
            logger.info(
                "The model is bigger than the maximum size per checkpoint %s "
                " and is going to be split in %s checkpoint shards. You can find "
                "where each parameters has been saved in the index located at (%s).",
                max_shard_size,
                str(len(shards)),
                save_index_file,
            )

    def close(self) -> None:
        """Save the tensors in the file."""
        self._save_to_file()



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli MixtralForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Iterator, List

import torch

from friendli.modules.quantizer.models.llama import LlamaHook
from friendli.modules.quantizer.schema.data import (
    HFQuantInput,
    HFTFQuantInputs,
    TFQuantInputs,
)


class MixtralHook(LlamaHook):
    """FP8QuantHook for MixtralForCausalLM."""

    def iter_tf_quant_inputs(
        self, model: torch.nn.Module
    ) -> Iterator[TFQuantInputs] | Iterator[HFTFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of MixtralForCausalLM."""
        for index, decoder_layer in enumerate(
            self.get_tf_blocks(model)  # type: ignore[union-attr, arg-type]
        ):
            self_attn = decoder_layer.self_attn
            block_sparse_moe = decoder_layer.block_sparse_moe
            moe_ff1_ff_gate_target_names = []
            for expert_idx in range(self.converter.num_experts):
                moe_ff1_ff_gate_target_names.extend(
                    [
                        f"{self.quantized_layer_prefix}{index}.block_sparse_moe.experts.{expert_idx}.w1",
                        f"{self.quantized_layer_prefix}{index}.block_sparse_moe.experts.{expert_idx}.w3",
                    ]
                )

            yield HFTFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                quant_inputs=[
                    HFQuantInput(
                        parent_module=self_attn,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.self_attn.q_proj",
                            f"{self.quantized_layer_prefix}{index}.self_attn.k_proj",
                            f"{self.quantized_layer_prefix}{index}.self_attn.v_proj",
                        ],
                        local_names=["q_proj", "k_proj", "v_proj"],
                    ),
                    HFQuantInput(
                        parent_module=self_attn,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.self_attn.o_proj",
                        ],
                        local_names=[
                            "o_proj",
                        ],
                    ),
                    # router
                    HFQuantInput(
                        parent_module=block_sparse_moe,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.block_sparse_moe.gate",
                        ],
                        local_names=["gate"],
                    ),
                    # ff1, ff_gate in each moe
                    HFQuantInput(
                        parent_module=block_sparse_moe.experts,
                        target_names=moe_ff1_ff_gate_target_names,
                        local_names=["w1", "w3"],
                    ),
                    # ff2 in each moe
                    HFQuantInput(
                        parent_module=block_sparse_moe.experts,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.block_sparse_moe.experts.{expert_idx}.w2"
                            for expert_idx in range(self.converter.num_experts)
                        ],
                        local_names=["w2"],
                    ),
                ],
            )



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli DbrxForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Dict, Iterator, List, Tuple, Type, cast

import torch
from torch.nn.modules import Module
from tqdm import tqdm
from transformers.models.dbrx.modeling_dbrx import DbrxBlock, DbrxConfig, DbrxExpertGLU

from friendli.modules.quantizer.base import FP8QuantHook
from friendli.modules.quantizer.schema.data import (
    HFQuantInput,
    HFTFQuantInputs,
    TFQuantInputs,
)


class DbrxLinearLayer(torch.nn.Module):
    """Custom FF2Proj layer for DbrxForCausalLM."""

    def __init__(self, weight: torch.nn.Parameter):
        """Initialize the DbrxLinearLayer."""
        super().__init__()
        self.weight = weight

    def forward(self, x: torch.Tensor, chunked_weight: torch.Tensor) -> torch.Tensor:
        """Forward pass for the DbrxLinearLayer."""
        return x.matmul(chunked_weight)


class CustomDbrxExpertGLU(DbrxExpertGLU):
    """Custom DbrxExpertGLU layer for DbrxForCausalLM.

    This layer is used to replace the DbrxExpertGLU layer in DbrxForCausalLM.
    For collecting input of the ff2 layer in each experts, we need to override the forward method.
    """

    def __init__(self, layer: DbrxExpertGLU, ffn_act_fn: Dict):
        """Initialize the CustomDbrxExpertGLU."""
        super().__init__(
            layer.hidden_size, layer.ffn_hidden_size, layer.moe_num_experts, ffn_act_fn
        )

        self.v1_linear = DbrxLinearLayer(layer.v1.detach())
        self.w1_linear = DbrxLinearLayer(layer.w1.detach())
        self.w2_linear = DbrxLinearLayer(layer.w2.detach())

    def forward(
        self,
        x: torch.Tensor,
        expert_w1: torch.Tensor,
        expert_v1: torch.Tensor,
        expert_w2: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass for the CustomDbrxExpertGLU."""
        gate_proj = self.w1_linear(x, expert_w1.t())
        up_proj = self.v1_linear(x, expert_v1.t())
        gate_proj = self.activation_fn(gate_proj)
        intermediate_states = gate_proj * up_proj
        down_proj = self.w2_linear(intermediate_states, expert_w2)
        return down_proj

    @staticmethod
    def from_layer(layer: DbrxExpertGLU, config: DbrxConfig) -> CustomDbrxExpertGLU:
        """Creates a CustomDbrxExpertGLU layer from a DbrxExpertGLU layer."""
        custom_layer = CustomDbrxExpertGLU(layer, config.ffn_config.ffn_act_fn)
        custom_layer.v1 = layer.v1
        custom_layer.w1 = layer.w1
        custom_layer.w2 = layer.w2
        return custom_layer


class DbrxHook(FP8QuantHook):
    """FP8QuantHook for DbrxForCausalLM."""

    def get_quantized_param_names(self, model: torch.nn.Module) -> List[str]:
        """Return the parameter names of quantized layers."""
        quantized_param_names = []
        for index in range(
            len(self.get_tf_blocks(model))  # type: ignore[union-attr, arg-type]
        ):
            quantized_param_names.extend(
                [
                    f"{self.quantized_layer_prefix}{index}.norm_attn_norm.attn.Wqkv.weight",
                    f"{self.quantized_layer_prefix}{index}.norm_attn_norm.attn.out_proj.weight",
                    f"{self.quantized_layer_prefix}{index}.ffn.router.layer.weight",
                    f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.v1",
                    f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.w1",
                    f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.w2",
                ]
            )
        return quantized_param_names

    def get_quantized_param_scale_names(self, model: torch.nn.Module) -> List[str]:
        """Return the parameter scale names of quantized layers."""
        quantized_param_scale_names = []
        for index in range(
            len(self.get_tf_blocks(model))  # type: ignore[union-attr, arg-type]
        ):
            quantized_param_scale_names.extend(
                [
                    f"{self.quantized_layer_prefix}{index}.norm_attn_norm.attn.Wqkv.weight_scale",
                    f"{self.quantized_layer_prefix}{index}.norm_attn_norm.attn.out_proj.weight_scale",
                    f"{self.quantized_layer_prefix}{index}.ffn.router.layer.weight_scale",
                    f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.v1_weight_scale",
                    f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.w1_weight_scale",
                    f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.w2_weight_scale",
                ]
            )
            quantized_param_scale_names.extend(
                [
                    f"{self.quantized_layer_prefix}{index}.norm_attn_norm.attn.Wqkv.in_scale",
                    f"{self.quantized_layer_prefix}{index}.norm_attn_norm.attn.out_proj.in_scale",
                    f"{self.quantized_layer_prefix}{index}.ffn.router.layer.in_scale",
                    f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.v1_in_scale",
                    f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.w1_in_scale",
                    f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.w2_in_scale",
                ]
            )
        return quantized_param_scale_names

    def pre_quantize(self, model: Module) -> torch.nn.Module:
        """Pre-quantization hook for DbrxForCausalLM."""
        for decoder_layer in tqdm(
            self.get_tf_blocks(model),
            desc="Pre-quantizing DbrxForCausalLM",
            unit="layer",
        ):
            cast(
                DbrxBlock, decoder_layer
            ).ffn.experts.mlp = CustomDbrxExpertGLU.from_layer(
                cast(DbrxBlock, decoder_layer).ffn.experts.mlp, self.converter.config
            )
        return model

    def post_quantize(self, model: Module) -> torch.nn.Module:
        """Post-quantization hook for DbrxForCausalLM."""
        for decoder_layer in tqdm(
            self.get_tf_blocks(model),
            desc="Post-quantizing DbrxForCausalLM",
            unit="layer",
        ):
            mlp = cast(DbrxBlock, decoder_layer).ffn.experts.mlp

            # ff1
            setattr(mlp, "v1_in_scale", mlp.v1_linear.in_scale)
            setattr(mlp, "v1_weight_scale", mlp.v1_linear.weight_scale)
            mlp.v1 = mlp.v1_linear.weight
            del mlp.v1_linear

            # ff_gate
            setattr(mlp, "w1_in_scale", mlp.w1_linear.in_scale)
            setattr(mlp, "w1_weight_scale", mlp.w1_linear.weight_scale)
            mlp.w1 = mlp.w1_linear.weight
            del mlp.w1_linear

            # ff2
            setattr(mlp, "w2_in_scale", mlp.w2_linear.in_scale)
            setattr(mlp, "w2_weight_scale", mlp.w2_linear.weight_scale)
            mlp.w2 = mlp.w2_linear.weight
            del mlp.w2_linear
        return model

    def get_tf_blocks(self, model: torch.nn.Module) -> List[torch.nn.Module]:
        """Returns the transformer blocks in DbrxForCausalLM."""
        return model.transformer.blocks

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in DbrxForCausalLM."""
        return (
            torch.nn.Linear,
            DbrxLinearLayer,
        )

    def iter_tf_quant_inputs(
        self, model: torch.nn.Module
    ) -> Iterator[TFQuantInputs] | Iterator[HFTFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of DbrxForCausalLM."""
        for index, decoder_layer in enumerate(
            self.get_tf_blocks(model)  # type: ignore[union-attr, arg-type]
        ):
            self_attn = cast(DbrxBlock, decoder_layer).norm_attn_norm.attn
            mlp = cast(DbrxBlock, decoder_layer).ffn.experts.mlp

            yield HFTFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                quant_inputs=[
                    HFQuantInput(
                        parent_module=self_attn,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.norm_attn_norm.attn.Wqkv",
                        ],
                        local_names=["Wqkv"],
                    ),
                    HFQuantInput(
                        parent_module=self_attn,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.norm_attn_norm.attn.out_proj",
                        ],
                        local_names=[
                            "out_proj",
                        ],
                    ),
                    HFQuantInput(
                        parent_module=cast(DbrxBlock, decoder_layer).ffn.router,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.ffn.router.layer",
                        ],
                        local_names=["layer"],
                    ),
                    HFQuantInput(
                        parent_module=mlp,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.w1_linear",
                            f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.v1_linear",
                        ],
                        local_names=["w1_linear", "v1_linear"],
                    ),
                    HFQuantInput(
                        parent_module=mlp,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.ffn.experts.mlp.w2_linear"
                        ],
                        local_names=["w2_linear"],
                    ),
                ],
            )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Quantizer Base."""

from __future__ import annotations

import os
from abc import ABC, abstractmethod
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any, Dict, Iterator, List, Tuple, Type, Union, cast

import datasets  # type: ignore[import]
import huggingface_hub  # type: ignore[import]
import numpy as np
import torch
from torch.nn.modules import Module
from tqdm import tqdm

from friendli.enums import (
    QuantDatasetFormat,  # TODO: move this to friendli/modules/converter/enums.py
)
from friendli.enums import ModelDataType
from friendli.errors import NotSupportedQuantConfigError
from friendli.logging import logger
from friendli.modules.converter.base import DECODER_PREFIX, OneOfConverter
from friendli.modules.converter.interface import ModelConversionInterface
from friendli.modules.converter.schema import ConvertInfo
from friendli.modules.converter.utils import get_tokenizer, get_torch_data_type
from friendli.modules.quantizer.layers import WeightActQuantizedLinearLayer
from friendli.modules.quantizer.schema.config import OneOfQuantConfig
from friendli.modules.quantizer.schema.data import (
    HFTFQuantInputs,
    ModuleName,
    TFQuantInputs,
    TFQuantResults,
    WeightActQuantResult,
)
from friendli.modules.quantizer.utils import (
    collect_stats,
    offload_module_sequence,
    safe_load_datasets,
    send_model_to_device,
)


class AbstractQuantHook(ABC):
    """Quantization Hook for a specific model architecture."""

    def __init__(self, quant_config: Dict[str, Any], converter: OneOfConverter):
        """Initialize the Quantization Hook.

        Args:
            quant_config: Quantization configuration.
            converter (OneOfConverter): Converter for a specific model architecture.
        """
        self.quant_config = quant_config
        self.converter = converter

    @abstractmethod
    def get_tf_blocks(self, model: torch.nn.Module) -> List[torch.nn.Module]:
        """Returns the transformer blocks."""

    @abstractmethod
    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the type of linear layer (etc. qkv, linear layer) in transformer block."""

    @abstractmethod
    def iter_tf_quant_inputs(
        self, model: torch.nn.Module
    ) -> Union[Iterator[TFQuantInputs], Iterator[HFTFQuantInputs]]:
        """Returns the layers which should be quantized in transformer blocks."""

    @abstractmethod
    def get_quant_result(
        self,
        quant_inputs: TFQuantInputs,
        **kwargs: Any,
    ) -> TFQuantResults:
        """Returns the quantization result of the layer."""

    @property
    @abstractmethod
    def quantized_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for quantized layers."""

    @property
    @abstractmethod
    def modified_layers_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for modified layers."""

    @property
    def quantized_layer_prefix(self) -> str:
        """Returns the prefix of the transformer block name."""
        return self.converter.decoder_layer_prefix

    @property
    def quantized_param_names(self) -> List[str]:
        """Return the parameter names of quantized layers."""
        param_names = []
        for i in range(self.converter.decoder_layer_num):
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            param_names.append(f"{converted_prefix}attn/c_attn/weight:0")
            param_names.append(f"{converted_prefix}attn/c_proj/weight:0")
            param_names.append(f"{converted_prefix}mlp/c_fc/weight:0")
            param_names.append(f"{converted_prefix}mlp/c_proj/weight:0")

        return param_names


class AbstractQuantizer(ABC):
    """Abstract Quantizer for a specific model architecture."""

    def __init__(
        self,
        hook: AbstractQuantHook,
        config: OneOfQuantConfig,
        converter: OneOfConverter,
    ):
        """Initialize the Quantizer.

        Args:
            hook (AbstractQuantHook): Quantization Hook for a specific model architecture
            config (CommonQuantConfig): Quantization configuration.
            converter (OneOfConverter): Converter for a specific model architecture.

        """
        self.hook = hook
        self.quant_config = config
        self.converter = converter

    @abstractmethod
    def get_calib_dataset(
        self,
    ) -> datasets.Dataset:
        """Get calibration dataset."""

    @abstractmethod
    def pre_quantize(
        self,
        model: torch.nn.Module,
    ) -> None:
        """Pre-procedure that should be called before quantize() is called."""

    @abstractmethod
    def quantize(
        self,
        model: torch.nn.Module,
    ) -> torch.nn.Module:
        """Setting Quantizer from config and Quantize model."""


class CommonQuantizer(AbstractQuantizer, ModelConversionInterface):
    """Common Quantizer."""

    def check_config(self) -> None:
        """Check if the quantization config is valid."""
        self.converter.check_config()
        calibration_dataset_config = self.quant_config.calibration_dataset
        data_path_or_name = calibration_dataset_config.path_or_name
        percentile = self.quant_config.percentile
        if percentile <= 0 or percentile > 100:
            raise NotSupportedQuantConfigError(
                invalid_option=str(percentile),
                valid_options=["0 < percentile <= 100"],
            )
        if not os.path.exists(data_path_or_name):
            data_name = data_path_or_name.split(":")[0]
            if data_name not in (
                data.id for data in huggingface_hub.list_datasets(search=data_name)
            ):
                raise NotSupportedQuantConfigError(
                    invalid_option=data_name,
                    valid_options=["datasets on the huggingface hub", "local path"],
                )
        else:
            if calibration_dataset_config.format not in QuantDatasetFormat:
                raise NotSupportedQuantConfigError(
                    invalid_option=calibration_dataset_config.format,
                    valid_options=list(QuantDatasetFormat),
                )
        try:
            torch.device(self.quant_config.device)
        except ValueError as err:
            raise NotSupportedQuantConfigError(
                invalid_option=self.quant_config.device,
                valid_options=["cpu", "cuda"],
            ) from err

    def get_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Get List of the convert informations for the model."""
        convert_info_list = self.converter.get_convert_info_list()
        new_convert_info_list = []
        for convert_info in convert_info_list:
            if convert_info.converted_name in self.hook.quantized_param_names:
                continue
            new_convert_info_list.append(convert_info)

        return (
            new_convert_info_list
            + self.hook.quantized_convert_info_list
            + self.hook.modified_layers_convert_info_list
        )

    def get_attributes(self) -> Dict[str, Any]:
        """Return the attributes of the converted model."""
        return self.converter.get_attributes()

    @contextmanager
    def _try_offload_model(self, model: torch.nn.Module):
        if not self.quant_config.offload:
            logger.info("Offloading not enabled. Skipping.")
            model.to(self.quant_config.device)
            yield
        else:
            logger.info("Offloading enabled.")
            tf_blocks = self.hook.get_tf_blocks(model)
            send_model_to_device(model, self.quant_config.device, exclude=tf_blocks)
            with offload_module_sequence(tf_blocks, self.quant_config.device):
                yield

    def convert(
        self,
        model: torch.nn.Module,
        convert_info_list: List[ConvertInfo],
        save_numpy_format: bool = True,
    ) -> Generator[Tuple[str, Union[np.ndarray, torch.Tensor]], None, None]:
        """Convert Huggingface Model to Friendli format(.h5).

        Args:
            model (torch.nn.Module): Huggingface model.
            state_dict (Dict[str, torch.Tensor]):
                Dictionary of mapping of tensor name to tensor
            convert_info_list (List[ConvertInfo]):
                Dictionary of mapping converted params name to conversion functions.
            save_numpy_format (bool, optional): Save the converted tensor in numpy format.
                                                Defaults to True.
        """
        self.pre_quantize(model)
        model = self.quantize(model)
        yield from self.converter.convert(model, convert_info_list, save_numpy_format)


class FP8QuantHook(AbstractQuantHook):
    """Quantization Hook for FP8Quantizer."""

    def pre_quantize(self, model: Module) -> torch.nn.Module:  # type: ignore[]
        """Pre-procedure that should be called before quantize() is called in FP8Quantizer."""
        return model

    def post_quantize(self, model: Module) -> torch.nn.Module:
        """Post-procedure that should be called after quantize() is called in FP8Quantizer."""
        return model

    def get_quant_result(
        self, quant_inputs: TFQuantInputs, **kwargs: Any
    ) -> TFQuantResults:
        """Returns the quantization result of the layer."""
        raise NotImplementedError

    def get_quantized_param_names(self, model: torch.nn.Module) -> List[str]:
        """Return the parameter names of quantized layers."""
        quantized_param_names = []
        for tf_quant_input in self.iter_tf_quant_inputs(model):
            assert isinstance(tf_quant_input, HFTFQuantInputs)
            for quant_input in tf_quant_input.quant_inputs:
                for target_name in quant_input.target_names:
                    quantized_param_names.append(f"{target_name}.weight")
        return quantized_param_names

    def get_quantized_param_scale_names(self, model):
        """Return the parameter scale names of quantized layers."""
        quantized_param_scale_names = []
        for tf_quant_input in self.iter_tf_quant_inputs(model):
            assert isinstance(tf_quant_input, HFTFQuantInputs)
            for quant_input in tf_quant_input.quant_inputs:
                for target_name in quant_input.target_names:
                    quantized_param_scale_names.append(f"{target_name}.weight_scale")
                    quantized_param_scale_names.append(f"{target_name}.in_scale")
        return quantized_param_scale_names

    @property
    def quantized_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for quantized layers."""
        raise NotImplementedError

    @property
    def modified_layers_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for modified layers."""
        raise NotImplementedError


class FP8Quantizer(CommonQuantizer):
    """FP8Quantizer for huggingface format.

    This quantizer supports per-tensor weight-activation quantization by
    using calibration dataset. It adds quantization scale, and quantized
    parameter to the checkpoint, while preserves parameter shape, and name
    in huggingface checkpoint.
    """

    def get_calib_dataset(self) -> datasets.Dataset:
        """Get calibration dataset."""
        data_cfg = self.quant_config.calibration_dataset
        tokenizer = get_tokenizer(self.converter.config.name_or_path)
        dataset = safe_load_datasets(data_cfg)

        dataset = (
            dataset.shuffle(self.quant_config.seed)
            .select(range(data_cfg.num_samples))
            .select_columns([data_cfg.lookup_column_name])
        )

        encoded_dataset = tokenizer(
            dataset[data_cfg.lookup_column_name],
            return_tensors="pt",
            padding=True,
            truncation=True,
            max_length=data_cfg.max_length,
        )
        return encoded_dataset["input_ids"]

    def get_convert_info_list(self) -> List[ConvertInfo]:
        """Not used in FP8Quantizer."""
        return []

    def pre_quantize(self, model: Module) -> None:
        """Not used in FP8Quantizer."""
        return None

    def _get_weight_act_quantize_results(
        self,
        model: torch.nn.Module,
        names: List[ModuleName],
        max_input_stats: Dict[ModuleName, torch.Tensor],
    ) -> List[WeightActQuantResult]:
        """Get the quantization scales and quantized_weight for a specific layer."""
        assert (
            self.quant_config.quant_dtype == ModelDataType.FP8_E4M3
        ), "currently support fp8_e4m3"
        max_val = 448.0
        min_val = -448.0

        input_max = torch.concat([max_input_stats[name] for name in names])
        target_weights = [model.get_submodule(name).weight for name in names]
        target_weight = torch.concat(target_weights)

        act_scale = float(input_max.detach().abs().max().item()) / float(max_val)
        weight_scale = float(target_weight.detach().abs().max().item()) / float(max_val)

        q_weights = [
            (
                (weight.detach().float() / weight_scale)
                .clip(min_val, max_val)
                .to(torch.float8_e4m3fn)
                .view(torch.int8)
                .to("cpu")
            )
            for weight in target_weights
        ]
        return [
            WeightActQuantResult(
                name,
                quant_dtype=self.quant_config.quant_dtype,
                act_scale=torch.tensor(act_scale, dtype=torch.float32),
                weight_scale=torch.tensor(weight_scale, dtype=torch.float32),
                q_weight=q_weight,
                q_group_size=-1,
                zero_point=torch.tensor(0.0),
            )
            for name, q_weight in zip(names, q_weights)
        ]

    @torch.no_grad()
    def quantize(
        self,
        model: torch.nn.Module,
    ) -> torch.nn.Module:
        """Quantize model to lower data type. Currently supports FP8."""
        # pylint: disable=too-many-locals
        dataset = self.get_calib_dataset()
        model.eval()
        with self._try_offload_model(model):
            max_input_stats, _ = collect_stats(
                model,
                self.quant_config.device,
                dataset,
                cast(FP8QuantHook, self.hook).get_linear_layer_types(),
                percentile=self.quant_config.percentile,
                tqdm_desc="Collecting stats for Static Quantization.",
                batch_size=32,
            )
            for tf_quant_input in tqdm(
                self.hook.iter_tf_quant_inputs(model),
                total=len(self.hook.get_tf_blocks(model)),
                desc="Qunatize",
                unit="layer",
            ):
                assert isinstance(tf_quant_input, HFTFQuantInputs)
                for quant_input in tf_quant_input.quant_inputs:
                    parent_module, local_names, names = (
                        quant_input.parent_module,
                        quant_input.local_names,
                        quant_input.target_names,
                    )

                    if isinstance(parent_module, torch.nn.ModuleList):
                        # For MoE models with seperate expert layers
                        parent_modules_w_local_name = []
                        for p_module in parent_module:
                            for local_name in local_names:
                                parent_modules_w_local_name.append(
                                    (p_module, local_name)
                                )

                        layers = [
                            p_module.get_submodule(local_name)
                            for p_module, local_name in parent_modules_w_local_name
                        ]

                        quant_results = self._get_weight_act_quantize_results(
                            model,
                            names,
                            max_input_stats,
                        )
                        q_layers = [
                            WeightActQuantizedLinearLayer.from_layer(
                                layer, quant_result
                            )
                            for layer, quant_result in zip(layers, quant_results)
                        ]
                        for (p_module, local_name), q_layer in zip(
                            parent_modules_w_local_name, q_layers
                        ):
                            setattr(p_module, local_name, q_layer)

                    else:
                        layers = [
                            parent_module.get_submodule(local_name)
                            for local_name in local_names
                        ]
                        quant_results = self._get_weight_act_quantize_results(
                            model,
                            names,
                            max_input_stats,
                        )
                        q_layers = [
                            WeightActQuantizedLinearLayer.from_layer(
                                layer, quant_result
                            )
                            for layer, quant_result in zip(layers, quant_results)
                        ]
                        for local_name, q_layer in zip(local_names, q_layers):
                            setattr(parent_module, local_name, q_layer)

        return model

    def convert(  # type: ignore[override]
        self,
        model: torch.nn.Module,
        convert_info_list: List[ConvertInfo],
        save_numpy_format: bool = False,
    ) -> Generator[Tuple[str, Union[torch.Tensor, np.ndarray]], None, None]:
        """Convert Huggingface Model to Friendli format(.h5).

        Args:
            model (torch.nn.Module): Huggingface model.
            state_dict (Dict[str, torch.Tensor]):
                Dictionary of mapping of tensor name to tensor
            convert_info_list (List[ConvertInfo]):
                Dictionary of mapping converted params name to conversion functions.
                It will be depreciated.
            save_numpy_format (bool, optional): Save the converted tensor in numpy format.
                                                It will be depreciated.
        """
        model = cast(FP8QuantHook, self.hook).pre_quantize(model)
        model = self.quantize(model)
        model = cast(FP8QuantHook, self.hook).post_quantize(model)
        state_dict: Dict[str, torch.Tensor] = model.state_dict()

        quantized_param_names = cast(FP8QuantHook, self.hook).get_quantized_param_names(
            model
        )
        quantized_param_names.extend(
            cast(FP8QuantHook, self.hook).get_quantized_param_scale_names(model)
        )

        with tqdm(total=len(state_dict), desc="Converting", unit="tensor") as pbar:
            for param_name, param in state_dict.items():
                if param_name not in quantized_param_names:
                    param = param.to(get_torch_data_type(self.converter.data_type))
                yield param_name, param
                pbar.update()



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli MPTForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Iterator, List, Tuple, Type

import torch

from friendli.modules.quantizer.base import FP8QuantHook
from friendli.modules.quantizer.schema.data import (
    HFQuantInput,
    HFTFQuantInputs,
    TFQuantInputs,
)


class MPTHook(FP8QuantHook):
    """FP8QuantHook for MPTForCausalLM."""

    def get_tf_blocks(self, model: torch.nn.Module) -> List[torch.nn.Module]:
        """Returns the transformer blocks in MPTForCausalLM."""
        return model.transformer.blocks

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in MPTForCausalLM."""
        return (torch.nn.Linear,)

    def iter_tf_quant_inputs(
        self, model: torch.nn.Module
    ) -> Iterator[TFQuantInputs] | Iterator[HFTFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of MPTForCausalLM."""
        for index, decoder_layer in enumerate(
            self.get_tf_blocks(model)  # type: ignore[union-attr, arg-type]
        ):
            self_attn = decoder_layer.attn
            mlp = decoder_layer.ffn

            yield HFTFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                quant_inputs=[
                    HFQuantInput(
                        parent_module=self_attn,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.attn.Wqkv",
                        ],
                        local_names=["Wqkv"],
                    ),
                    HFQuantInput(
                        parent_module=self_attn,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.attn.out_proj",
                        ],
                        local_names=[
                            "out_proj",
                        ],
                    ),
                    HFQuantInput(
                        parent_module=mlp,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.ffn.up_proj",
                        ],
                        local_names=["up_proj"],
                    ),
                    HFQuantInput(
                        parent_module=mlp,
                        target_names=[
                            f"{self.quantized_layer_prefix}{index}.ffn.down_proj"
                        ],
                        local_names=["down_proj"],
                    ),
                ],
            )



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli Cohere Checkpoint Converter."""


from __future__ import annotations

from typing import cast

from transformers import CohereConfig  # type: ignore[import]

from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.modules.converter.base import FP8OnlyConverter
from friendli.modules.converter.interface import RotaryEmbeddingConversionInterface
from friendli.modules.converter.schema import ConvertInfo


class CohereForCausalLMConverter(FP8OnlyConverter, RotaryEmbeddingConversionInterface):
    """CohereForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if LLaMA architectures' config can be converted to Friendli format."""
        super().check_config()
        try:
            if cast(CohereConfig, self.config).hidden_act not in ["silu"]:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'hidden_act={cast(CohereConfig, self.config).hidden_act}'",
                    valid_options=["silu"],
                )
            if not cast(CohereConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=False'",
                    valid_options=[True],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    @property
    def model_type(self) -> str:
        """Model type."""
        return "cohere"

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before LLaMA's transformer block number."""
        return "model.layers."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in LLaMA."""
        return cast(CohereConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in LLaMA."""
        return cast(CohereConfig, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in LLaMA."""
        return cast(CohereConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in LLaMA."""
        config = cast(CohereConfig, self.config)
        if config.num_key_value_heads is None:
            return self.decoder_num_attention_heads
        return config.num_key_value_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of LLaMA."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate size of the linear layer in LLaMA MLP."""
        return self.config.intermediate_size

    @property
    def rotary_dim(self) -> int:
        """The rotary embedding dimension of LLaMA."""
        return self.decoder_head_size

    @property
    def rotary_emb_base(self) -> float:
        """The rotary embedding base of LLaMA."""
        return cast(CohereConfig, self.config).rope_theta
